(function () {

var HTE_VERSION = "HTE 1.0";

var HTE_ACTION = {
    HTE_CLIENT_ACTION_CONNECTED               : 1,
    HTE_CLIENT_ACTION_UPDATE                  : 2,
    HTE_CLIENT_ACTION_SET_CURSOR              : 3,
    HTE_CLIENT_ACTION_GET_ENTRY               : 4,
    HTE_CLIENT_ACTION_GET_YESNO               : 5,
    HTE_CLIENT_ACTION_SHOW_FATAL_ERROR        : 6,
    HTE_CLIENT_ACTION_STATUS_UPDATE           : 7,
    HTE_CLIENT_ACTION_SET_SIZE                : 8,
    HTE_HOST_ACTION_REFRESH                   : 4000,
    HTE_HOST_ACTION_CONNECT_HOST              : 4001,
    HTE_HOST_ACTION_SET_SIZE                  : 4002,
    HTE_HOST_ACTION_ONKEY                     : 4003,
    HTE_HOST_ACTION_ON_MOUSE_DOWN             : 4004,
    HTE_HOST_ACTION_ON_MOUSE_UP               : 4005,
 	HTE_HOST_ACTION_ON_MOUSE_MOVE             : 4006,
    HTE_HOST_ACTION_GET_ENTRY_RESPONSE        : 4007,
    HTE_HOST_ACTION_GET_YESNO_RESPONSE        : 4008,
    HTE_HOST_ACTION_ACK_FATAL_ERROR           : 4009
};

var ANSI_ATTR = {
	NONE		:	0x00000000,
	BOLD		:	0x00100000,
	UNDERLINE	:	0x00200000,
	BLINK		:	0x00400000,
	REVERSE		:	0x00800000
};

var TEColors = [
	"#000000", // black
	"#000080", // lt-blue
	"#008000", // lt-green
	"#008080", // lt-cyan
	"#800000", // lt-red
	"#800080", // lt-magenta
	"#808080", // gold
	"#c1c1c1", // lt-gray
	"#808080", // gray
	"#0000ff", // blue
	"#00ff00", // green
	"#00ffff", // cyan
	"#ff0000", // red
	"#ff00ff", // magenta
	"#ffff00", // yellow
	"#ffffff"  // white
];

var TE_Highlight_Modes = {
	BLOCK: 0,
	LINE: 1
};

var TE_KeyCode = {
	F1: 63236,
	SF1: 61769,
	CF1: 61789,
	CSF1: 61829,
};

var IB_KeyCode = {
	F1: 112,
	F12: 123
};

var IB_TO_TE = {
	33: 0xf128, 34: 0xf127, 35: 0xf72b, 36: 0xf729, // pgup, pgdown, end, home
	37: 0xf117, 38: 0xf114, 39: 0xf116, 40: 0xf115, // left, up, right, down
	9: 9, 46: 61741, // tab, forward delete
	96: 61721, 97: 61722, 98: 61723, 99: 61724, 100: 61725, // KP0-KP9
	101: 61726, 102: 61727, 103: 61728, 104: 61729, 105: 61730,
	106: 61754, 107: 61755, 109: 61731, 110: 61733, 111: 61756 // star, plus, dash, dot, slash
};

var IB_TO_TE_C = {
	9: 61766, 33: 61746, 34: 61747, 35: 61739, 36: 61737, // tab, pgup, pgdown, end, home
	37: 61743, 38: 61760, 39: 61742, 40: 61761, // left, up, right, down
	111: 61767, 106: 61753, 109: 61751, 107: 61752 // slash, star, dash, plus
};

var IB_TO_TE_S = {
	9: 61745 // tab
};

window.HTEClientConsole = HTEClientConsole;

function HTEClientConsole (id) {
	var teConsole = this;
	var cols = 80;
	var lines = 25;
	var factor = 4;
	var emulate;
	var cellSizeX = 8 * factor;
	var cellSizeY = 20 * factor;
	var fontSize = 12 * factor;
	var cursorX = 0;
	var cursorY = 0;
	var cursorColor = "#00ff00";
	var cursorBlinkState = true;
	var cursorVisible = true;
	var cursorRestoreImage;
	var cursorTimer = -1;
	var blinkTimer = -1;
	var showBlinkCharacters = false;
	var blinkCharacters = {};
	var handleKeyEvents = true;
	var showStatusBar = false;
	var statusBarSepHeight = 9;
	var element = document.getElementById(id);
	var highlightMouseDown = false;
	var highlightRectShown = false;
	var highlightRect = Rect(NaN, NaN, NaN, NaN);
	var highlightMode = TE_Highlight_Modes.LINE;
	var highlightBG = 14;
	var highlightFG = 9;
	var includeTrailingSpaces = false;
	var vscreen = {};


	function pasty(event) {
		alert("pasty");
	}

	openTeChannel();

	function initConsole() {

		document.addEventListener("paste",function(e) {
			performPaste();
			e.preventDefault();
		});

		document.addEventListener("copy", function(e) {
			performCopy();
			e.preventDefault();
		});

		teConsole.getConfigValues(["emulate", "lines", "col", "CursorColor"], function (obj) {

			emulate = parseInt(obj.emulate, 10);

			if (emulate >= 24 && emulate <= 27) {
				showStatusBar = true;

				switch (emulate) {
					case 24:
						cols = 80;
						lines = 24;
						break;
					case 25:
						cols = 80;
						lines = 32;
						break;
					case 26:
						cols = 80;
						lines = 43;
						break;
					case 27:
						cols = 132;
						lines = 27;
						break;
				}
			}
			else if (emulate == 22) {
				showStatusBar = true;
				cols = 80;
				lines = 25;
			}
			else if (emulate == 23) {
				showStatusBar = true;
				cols = 132;
				lines = 27;
			}
			else {
				showStatusBar = false;
				lines = parseInt(obj.lines, 10);
				cols = parseInt(obj.col, 10);
			}

			cursorColor = TEColors[obj.CursorColor];

			WebFont.load({
				custom: {
					families: ["lucon"],
					urls: ["/css/console.css"]
				},
				fontactive: function (fontFamily, fvd) {
					if (fontFamily == "lucon") {
						initScreen();
						hteRequestRefresh();
					}
				}
			});
		});
	}

	function getCanvasCtx() {
		if (!getCanvasCtx.ctx)
			getCanvasCtx.ctx = element.getContext("2d");

		return getCanvasCtx.ctx;
	}

	function initScreen() {
		ctx = getCanvasCtx();

		element.width = cellSizeX * cols;

		if (showStatusBar) {
			element.height = cellSizeY * (lines + 1) + statusBarSepHeight;
		}
		else {
			element.height = cellSizeY * lines;
		}

		ctx.font = fontSize + "px Lucon";
		ctx.fillStyle = '#000000';
		ctx.fillRect(0, 0, element.width, element.height);

		if (showStatusBar) {
			var statusBarY = lines * cellSizeY;
			ctx.beginPath();
			ctx.strokeStyle = "#00ff00";
			ctx.lineWidth = statusBarSepHeight;
			ctx.moveTo(0, statusBarY);
			ctx.lineTo(element.width, statusBarY);
			ctx.stroke();
		}

		if (cursorTimer == -1)
			cursorTimer = setInterval(toggleCursor, 400);

		if (blinkTimer == -1)
			blinkTimer = setInterval(toggleBlinkCharacters, 400);

		closeDialog("#entryDialog");

		handleKeyEvents = true;
	}

	function drawChar(line, col, bg, fg, ch, showForeground) {
		var ctx = getCanvasCtx();

		var x = col * cellSizeX;
		var y = line * cellSizeY;

		if (showStatusBar && line == lines) {
			y += statusBarSepHeight;
		}

		ctx.fillStyle = TEColors[bg];
		ctx.fillRect(x, y, cellSizeX, cellSizeY);

		if (showForeground) {
			ctx.fillStyle = TEColors[fg];
			ctx.fillText(ch, x, y + fontSize);
		}
	}

	function drawAttr(attr, line, col, showForeground, highlighted) {
		var ctx = getCanvasCtx();

		var bg, fg, ch;

		if(!attr || !attr[0]) {
			attr = {};
			attr[0] = 0;
			attr[1] = 32;
			bg = highlightBG;
			fg = highlightFG;
			ch = 65;
		} else {
			bg = (attr[0] >> 28) & 0x0f;
			fg = (attr[0] >> 24) & 0x0f;
			ch = String.fromCharCode(attr[1] & ~(1 << 31));
		}

		var x = col * cellSizeX;
		var y = line * cellSizeY;

		var t;

		if (attr[0] & ANSI_ATTR.REVERSE) {
			t = bg;
			bg = fg;
			fg = t;
		}

		if(highlighted) {
			bg = highlightBG;
			fg = highlightFG;
			showForeground = true;
		}

		if (attr[0] & ANSI_ATTR.BLINK) {
			if (!blinkCharacters[line])
				blinkCharacters[line] = {};
			blinkCharacters[line][col] = attr;
		}
		else if (blinkCharacters[line] && blinkCharacters[line][col])
			delete(blinkCharacters[line][col]);

		drawChar(line, col, bg, fg, ch, showForeground);

		if (attr[0] & ANSI_ATTR.UNDERLINE) {
			ctx.beginPath();
			ctx.strokeStyle = ctx.fillStyle;
			ctx.lineWidth = 2;
			ctx.moveTo(x, y + cellSizeY);
			ctx.lineTo(x + cellSizeX, y + cellSizeY);
			ctx.stroke();
		}
	}

	function displayString(startX, y, bg, fg, str) {
		var i;
		var endY = lines;

		if (showStatusBar) {
			endY += 1;
		}

		if (startX >= cols || y >= endY) {
			return;
		}

		str = str.substr(0, cols - startX);

		for (i = 0; i < str.length; i++) {
			drawChar(y, startX + i, bg, fg, str[i], true);
		}
	}

	function displayCursorPosition(x, y) {
		var str = "";
		var statusBarY = lines;
		var startX;

		if (!showStatusBar) {
			return;
		}

		var str = "";

		if (x < 10)
			str += "0";
		if (x < 100)
			str += "0";

		str += x;
		str += "/";

		if (y < 10)
			str += "0";
		if (y < 100)
			str += "0";

		str += y;

		startX = cols - (str.length + 2);

		displayString(startX, statusBarY, 0, 3, str);
	}

	function displayStatusMessage(msg) {
		if (!showStatusBar) {
			return;
		}

		displayString(2, lines, 0, 3, msg);
	}

	function showCursor() {
		if (!cursorVisible || !cursorBlinkState)
			return false;

		var ctx = getCanvasCtx();

		cursorRestoreImage = ctx.getImageData(cursorX, cursorY, cellSizeX, cellSizeY);

		ctx.fillStyle = cursorColor;
		ctx.fillRect(cursorX, cursorY, cellSizeX, cellSizeY);

		cursorBlinkState = false;

		return true;
	}

	function hideCursor() {
		if (cursorBlinkState)
			return false;

		var ctx = getCanvasCtx();

		ctx.putImageData(cursorRestoreImage, cursorX, cursorY);

		cursorBlinkState = true;

		return true;
	}

	function setCursorPosition(x, y) {
		hideCursor();
		cursorX = x * cellSizeX;
		cursorY = y * cellSizeY;
		showCursor();

		displayCursorPosition(x + 1, y + 1);
	}

	function toggleCursor() {
		if (cursorBlinkState)
			showCursor();
		else
			hideCursor();
	}

	function clearHighlight(r) {
		for(var i = 0; i < r.length; i++) {
			displayRect(r[i], false);
		}
	}

	function displayRect(r, highlighted) {
		for(var y = r.top; y <= r.bottom; y++) {
			for(var x = r.left; x <= r.right; x++) {
				drawAttr(vscreen[y][x], y, x, true, highlighted);
			}
		}
	}

	function displayHighlight(r) {
		for(var i = 0; i < r.length; i++) {
			displayRect(r[i], true);
		}
	}

	function copyTextToClipboard(text) {
		navigator.clipboard.writeText(text);
	}

	function vscreenChar(x,y) {
		return vscreen[y][x][1] & ~(1 << 31);
	}

	function textFromRect(r) {
		var ret = "";
		var trailingSpaces = "";

		for(var y = r.top; y <= r.bottom; y++) {
			for(var x = r.left; x <= r.right; x++) {
				if(vscreen[y] && vscreen[y][x]) {
					var c = vscreenChar(x,y);
					if(c == 32) {
						trailingSpaces += " ";
					} else {
						ret += trailingSpaces;
						trailingSpaces = "";
						ret += String.fromCharCode(c);
					}
				} else {
					trailingSpaces += " ";
				}
			}

			if(includeTrailingSpaces) {
				ret += trailingSpaces;
				trailingSpaces = "";
			}

			if(y < r.bottom) {
				ret += "\n";
				trailingSpaces = "";
			}
		}

		return ret;
	}

	function getHighlightText(r) {
		if(highlightRectShown == false) {
			return "";
		}

		if(r.length == 1) {
			return textFromRect(r[0]);
		}

		var ret = "";

		for(var i=0; i < r.length; i++) {
			ret += textFromRect(r[i]);
			if(i < r.length-1) {
				ret += "\n";
			}
		}

		return ret;
	}

	function toggleBlinkCharacters() {
		for (var y in blinkCharacters)
			for (var x in blinkCharacters[y])
				drawAttr(blinkCharacters[y][x], y, x, showBlinkCharacters, false);

		showBlinkCharacters = !showBlinkCharacters;
	}

	function getASCIIStringFromInt8Array(bytes, len) {
		var str = "";

		for (var i = 0; i < len; i++) {
			str += String.fromCharCode(bytes[i]);
		}

		return str;
	}

	function getBodyOffset(bytes) {
		var i = 2;

		for (;;) {
			if (bytes[i - 2] == 0x0A && bytes[i - 1] == 0x0A) {
				for (;;) {
					if (bytes[i++] == 0x0A) {
						return i;
					}
				}
			}

			i++;
		}
	}

	function getArrayBufferOffset(buf, offset) {
		if (buf.slice)
			return buf.slice(offset);
		else {
			var newLen = buf.byteLength - offset;
			var newBuf = new ArrayBuffer(newLen);

			var src = new Int8Array(buf);
			var dst = new Int8Array(newBuf);

			for (var i = 0; i < newLen; i++) {
				dst[i] = src[i + offset];
			}

			return newBuf;
		}
	}

	function getHeaderFields(bytes) {
		var i = getBodyOffset(bytes);

		str = getASCIIStringFromInt8Array(bytes, i);

		return str.split("\n");
	}

	/* Send a message over the HTE channel.
	 *
	 * action - A numeric code that corresponds to an HTE action.
	 * fields - An optional array of strings that contains information about the message.
	 * data - An optional ArrayBuffer object that contains binary data.
	 *
	 * While the fields and data arguments are optional as per the communication
	 * protocol, each action has its own requirements. */
	this.hteSendMessage = function (action, fields, data) {
		var data;

		if (!_ws)
			return;

		if (!fields)
			fields = [];

		if (!data)
			data = new ArrayBuffer(0);

		if (typeof action !== "number") {
			console.error("Error (hteSendMessage): Argument [action] must be of type [number].");
			return;
		}

		if (typeof fields !== "object" || !(fields instanceof Array)) {
			console.error("Error (hteSendMessage): Argument [fields] must be an array.");
			return;
		}

		if (typeof data !== "object" || !(data instanceof ArrayBuffer)) {
			console.error("Error (hteSendMessage): Argument [data] must be an object of type [ArrayBuffer].");
			return;
		}

		fields.unshift(action);
		fields.unshift(HTE_VERSION);
		fields.push("\n");

		var hdr = fields.map(String).join("\n") + String(data ? data.byteLength : 0) + "\n";

		_ws.send(ArrayBuffer.combine(hdr.toUTF8(), data));
	};

	function hteRequestRefresh() {
		teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_REFRESH);
	}

	function hteSendKey(keyCode) {
		if(highlightRectShown && !highlightMouseDown) {
			clearHighlight(getHighlightRects());
			highlightRectShown = false;
		}
		var data = new ArrayBuffer(4);
		var ptr = new Int32Array(data);

		ptr[0] = keyCode;

		teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_ONKEY, [], data);
	}

	function hteSendMouseDown(button, x, y) {
		teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_ON_MOUSE_DOWN, [button, x, y], null);
	}

	function hteSendMouseUp(button, x, y) {
		teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_ON_MOUSE_UP, [button, x, y], null);
	}

	function Rect(left, top, right, bottom) {
		return { left: left, top: top, right: right, bottom: bottom };
	}

	function ScreenRectFromChar(r) {
		return Rect(cellSizeX * r.left, cellSizeY * r.top, cellSizeX * (r.right+1), cellSizeY * (r.bottom+1));
	}

	function dumpRect(r) {
		console.log("top = " + r.top + " left = " + r.left + " bottom = " + r.bottom + " right = " + r.right);
	}

	function dumpHighlightRects(r)
	{
		for(var i = 0; i < r.length; i++) {
			dumpRect(r[i]);
		}
	}

	function getHighlightRects() {
		var hr = {};

		Object.assign(hr, highlightRect);
		if(hr.top > hr.bottom || (hr.top == hr.bottom && hr.left > hr.right)) {
			var t;
			t = hr.bottom;
			hr.bottom = hr.top;
			hr.top = t;

			t = hr.right;
			hr.right = hr.left;
			hr.left = t;

		}

		var r = [];

		if(highlightMode == TE_Highlight_Modes.BLOCK) {
			r[0] = Rect(hr.left, hr.top, hr.right, hr.bottom);
		} else {
			var highlightLines = hr.bottom - hr.top + 1;
			var rectCount = highlightLines > 3 ? 3 : highlightLines;
			var count = 0;

			var numColumns = cols;

			if(highlightLines == 1) {
				r[count++] = Rect(hr.left, hr.top, hr.right, hr.bottom);
			}

			if(highlightLines > 1) {
				r[count++] = Rect(hr.left, hr.top, numColumns, hr.top);
				if(highlightLines > 2) {
					r[count++] = Rect(0, hr.top + 1, numColumns, hr.bottom - 1);
				}
				r[count++] = Rect(0, hr.bottom, hr.right, hr.bottom);
			}
		}

		return r;
	}

	function isHighlightChar(c) {
		if(!c) {
			return false;
		}
		return ((c > 63 && c < 91) || (c > 96 && c < 123) || c == 46 || (c > 47 && c < 58));
	}

	function highlightDoubleClick(x,y) {
		var c = vscreenChar(x,y);
		highlightRect.top = y;
		highlightRect.bottom = y;
		highlightRect.left = x;
		highlightRect.right = x;
		if(c && isHighlightChar(c)) {
			var left = x;
			var right = x;
			c = vscreenChar(left, y);
			while(left >= 0 && isHighlightChar(c)) {
				left--;
				c = vscreenChar(left >= 0 ? left : 0, y);
			}
			if(left >= 0) {
				left++;
			}
			highlightRect.left = left;

			c = vscreenChar(right, y);
			while(right < cols && isHighlightChar(c)) {
				right++;
				c = vscreenChar( right < cols ? right : cols-1, y);
			}
			if(right < cols) {
				right--;
			}
			highlightRect.right = right;
		}
		highlightRectShown = true;
		highlightMouseDown = false;
		displayHighlight(getHighlightRects());
	}

	function highlightBegin(x, y) {
		this.contenteditable = true;
		clearHighlight(getHighlightRects());
		highlightRect.top = y;
		highlightRect.left = x;
		highlightRect.bottom = y;
		highlightRect.right = x;
		highlightRectShown = false;
		highlightMouseDown = true;
	}

	function highlightUpdate(x,y) {
		if(highlightMouseDown) {
			if(highlightRectShown == false) {
				highlightRectShown = true;
			}
			clearHighlight(getHighlightRects());
			highlightRect.bottom = y;
			highlightRect.right = x;
			displayHighlight(getHighlightRects());
		}
	}

	function highlightEnd(x, y) {
		if(highlightRectShown == false) {
			highlightMouseDown = false;
			return;
		}
		highlightRect.bottom = y;
		highlightRect.right = x;
		highlightMouseDown = false;
		displayHighlight(getHighlightRects());
	}


	function performCopy()
	{
		if(highlightRectShown) {
			var text = getHighlightText(getHighlightRects());
			copyTextToClipboard(text);
		}
	}

	function performPaste()
	{
		clearHighlight(getHighlightRects());
		highlightRectShown = false;
		navigator.clipboard.readText().then(text => {
			for(var i = 0; i < text.length; i++) {
				hteSendKey(text.charCodeAt(i));
			}
		});
	}

	this.toolbarPaste = performPaste;
	this.toolbarCopy = performCopy;

	function openTeChannel() {
		var protocol = (window.location.protocol === "https:") ? "wss:" : "ws:";

		var port = "";

		if ((window.location.protocol == "https:" && window.location.port != 443) ||
			(window.location.protocol == "http:" && window.location.port != 80)) {
			port = ":" + window.location.port;
		}

		var url = protocol + "//" + window.location.hostname + port + "/get_te_channel";

		url += window.location.search;

		_ws = new WebSocket(url, "teCtrl");

		_ws.onopen = function () {
			$(document).on("keydown", function (e) {
				if (!handleKeyEvents)
					return;

				var keyCode = e.keyCode;
				var mappedKey;

				if (keyCode == 8 || // bs
					keyCode == 27) { // esc
					hteSendKey(keyCode);
					e.preventDefault();
				}
				else if (keyCode >= IB_KeyCode.F1 && keyCode <= IB_KeyCode.F12) {
					var fKey = keyCode - IB_KeyCode.F1;
					var offset;
					if (e.shiftKey) {
						if (e.ctrlKey)
							offset = TE_KeyCode.CSF1;
						else
							offset = TE_KeyCode.SF1;
					}
					else if (e.ctrlKey)
						offset = TE_KeyCode.CF1;
					else
						offset = TE_KeyCode.F1;

					hteSendKey(offset + fKey);
				}
				else {
					if (e.ctrlKey && (keyCode >= '@'.charCodeAt(0) && keyCode <= '_'.charCodeAt(0))) {
						mappedKey = keyCode & ~0x60;
					} else if(e.ctrlKey && keyCode == 45) {
						performCopy();
						e.preventDefault();
					} else if(e.shiftKey && keyCode == 45) {
						performPaste();
						e.preventDefault();
					} else if(e.metaKey) {
						if(keyCode == 67) {
							performCopy();
							e.preventDefault();
						} else if(keyCode == 86) {
							performPaste();
							e.preventDefault();
						}
					} else {
						if (e.shiftKey)
							mappedKey = IB_TO_TE_S[keyCode];
						else if (e.ctrlKey)
							mappedKey = IB_TO_TE_C[keyCode];
						else
							mappedKey = IB_TO_TE[keyCode];
					}

					if (mappedKey) {
						hteSendKey(mappedKey);
						e.preventDefault();
					}
				}
			});

			$(document).on("keypress", function (e) {
				if (!handleKeyEvents)
					return;

				hteSendKey(e.which || e.keyCode);

				e.preventDefault();
			});

			function mouseConvertCoordinates(e) {
				if(e.pageX < 1) {
					e.pageX = 1;
				}
				if(e.pageY < 1) {
					e.pageY = 1;
				}
				var x = ((e.pageX / $(document).width()) * 800).toFixed(0);
				var y = ((e.pageY / $(document).height()) * 600).toFixed(0);
				var charX = Math.floor(e.pageX / $(document).width() * cols);
				var charY = Math.floor(e.pageY / $(document).height() * lines);

				return { x: x, y: y, charX: charX, charY : charY };
			}

			$(document).on("dblclick", function (e) {
				var c = mouseConvertCoordinates(e);
				if(highlightRectShown) {
					clearHighlight(getHighlightRects());
					highlightRectShown = false;
				}
				highlightMouseDown = false;

				highlightDoubleClick(c.charX,c.charY);
			});

			$(document).on("mousedown", function (e) {
				if(e.button != 0) {
					return;
				}
				var c = mouseConvertCoordinates(e);

				hteSendMouseDown(0, c.x, c.y);
				highlightBegin(c.charX,c.charY);
			});

			$(document).on("mousemove", function (e) {
				var c = mouseConvertCoordinates(e);
				highlightUpdate(c.charX,c.charY);
			});

			$(document).on("mouseup", function (e) {
				if(e.button != 0) {
					return;
				}
				var c = mouseConvertCoordinates(e);

				hteSendMouseUp(0, c.x, c.y);
				highlightEnd(c.charX,c.charY);
			});
		};

		_ws.onmessage = function (msg) {
			var reader = new FileReader();

			reader.addEventListener("loadend", function () {
				var bytes = new Int8Array(reader.result);
				var fields = getHeaderFields(bytes);
				var actionCode = parseInt(fields[1], 10);

				if (actionCode === HTE_ACTION.HTE_CLIENT_ACTION_CONNECTED) {
					var isConnected = parseInt(fields[2], 10);

					if (isConnected) {
						initConsole();
					}
				}
				else if (actionCode === HTE_ACTION.HTE_CLIENT_ACTION_SET_CURSOR) {
					var m = fields[2].match(/^([0-9]+):([0-9]+)$/);

					if (m != null) {
						var x = parseInt(m[1], 10);
						var y = parseInt(m[2], 10);

						cursorVisible = parseInt(fields[3], 10);

						setCursorPosition(x, y);
					}
				}
				else if (actionCode == HTE_ACTION.HTE_CLIENT_ACTION_STATUS_UPDATE) {
					displayStatusMessage(fields[2]);
				}
				else if (actionCode == HTE_ACTION.HTE_CLIENT_ACTION_SET_SIZE) {
					lines = parseInt(fields[2], 10);
					cols = parseInt(fields[3], 10);

					initScreen();
				}
				else if (actionCode === HTE_ACTION.HTE_CLIENT_ACTION_UPDATE) {
					var screenFragment = new Int32Array(getArrayBufferOffset(reader.result, getBodyOffset(bytes)));

					var m = fields[2].match(/^([0-9]+):([0-9]+):([0-9]+):([0-9]+)$/);

					var left = parseInt(m[1], 10);
					var bottom = parseInt(m[2], 10);
					var right = parseInt(m[3], 10);
					var top = parseInt(m[4], 10);
					var x, y;
					var pos = 0;

					hideCursor();

					for (y = top; y <= bottom; y++) {
						if(!vscreen[y]) {
							vscreen[y] = {};
						}
						for (x = left; x <= right; x++) {
							var attr = [screenFragment[pos], screenFragment[pos + 1]];
							drawAttr(attr, y, x, true, false);
							vscreen[y][x] = attr;
							pos += 2;
						}
					}

					if(highlightRectShown) {
						clearHighlight(getHighlightRects());
						displayHighlight(getHighlightRects());
					}
				}
				else if (actionCode == HTE_ACTION.HTE_CLIENT_ACTION_GET_ENTRY) {
					handleKeyEvents = false;

					var isPassword = parseInt(fields[4], 10);

					var title = decodeURIComponent(fields[2]);
					var msg = decodeURIComponent(fields[3]);

					getEntry(title, msg, isPassword, function (status, msg) {

						teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_GET_ENTRY_RESPONSE, [
							encodeURIComponent(msg),
							status
						]);

						handleKeyEvents = true;
					});
				}
				else if (actionCode == HTE_ACTION.HTE_CLIENT_ACTION_GET_YESNO) {
					handleKeyEvents = false;

					var title = decodeURIComponent(fields[2]);
					var msg = decodeURIComponent(fields[3]);

					getYesNoQuestionResponse(title, msg, function (status) {

						teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_GET_YESNO_RESPONSE, [ status ]);

						handleKeyEvents = true;
					});
				}
				else if (actionCode == HTE_ACTION.HTE_CLIENT_ACTION_SHOW_FATAL_ERROR) {
					handleKeyEvents = false;

					var msg = decodeURIComponent(fields[3]);

					showFatalError(msg, function () {
						teConsole.hteSendMessage(HTE_ACTION.HTE_HOST_ACTION_ACK_FATAL_ERROR);

						handleKeyEvents = true;
					});
				}
				else {
				}
			});

			reader.readAsArrayBuffer(msg.data);
		};

		_ws.onclose = function (e) {
			if (!e.wasClean) {
				showFatalError("Connection closed by host.");
			}

			_ws = null;
		};

		_ws.onerror = function (err) {
            console.log(err);
		};
	}
};

HTEClientConsole.prototype.getConfigValues = function (keys, cb) {
	if (!keys || !keys.length) {
		alert("Must specify keys to fetch from host.");
		return;
	}

	if (!cb) {
		alert("Must provide callback function.");
		return;
	}

	$.ajax({
		url: "/get_config_values" + window.location.search,
		type: "POST",
		contentType: "application/json",
		dataType: "json",
		data: JSON.stringify(keys)
	})
	.done(function (res) {
		cb(res.content);
	});
};

function utf8_size(code)
{
	var bits = 31;

	if (code <= 127)
		return 1;

	while (!(code & (1 << bits))) {
		bits--;
	}

	return Math.ceil(bits / 5);
}

function appendUTF8ForCode(buffer, code)
{
	var n_bytes;
	var bytes;
	var bitsLeft;
	var i, j;

	if (!buffer || !(buffer instanceof Array)) {
		alert("Error (appendUTF8ForCode): Argument [buffer] must be of type [array].");
		return;
	}

	if (code <= 127) {
		buffer.push(code);
		return;
	}

	n_bytes = utf8_size(code);
	bytes = new Array(n_bytes);

	for (i = n_bytes - 1; i >= 0; i--) {
		var byte = 0;

		bitsLeft = 7;

		if (i == 0) {
			for (j = 0; j < n_bytes; j++) {
				byte |= (1 << (7 - j));
				bitsLeft--;
			}

			bitsLeft--;
		}
		else {
			byte |= 0x80;
			bitsLeft = 5;
		}

		j = 0;

		if (code > 0) { // still have bits left
			while (bitsLeft >= 0) { // remaining bits in this byte
				if (code & 1) {
					byte |= (1 << j);
				}

				j++;
				bitsLeft--;
				code >>= 1;
			}
		}

		bytes[i] = byte;
	}

	for (i = 0; i < bytes.length; i++)
		buffer.push(bytes[i]);
}

String.prototype.toUTF8 = function () {
	var i;
	var code;
	var bytes = [];
	var buf;
	var u8;

	for (i = 0; i < this.length; i++) {
		code = this.charCodeAt(i);

		if (code >= 0xD7FF && code <= 0xE000) {
			var code2 = this.charCodeAt(++i);

			code -= 0xD800;
			code2 -= 0xDC00;

			code <<= 10;

			code = 0x10000 + (code | code2);
		}

		appendUTF8ForCode(bytes, code);
	}

	buf = new ArrayBuffer(bytes.length);
	v8 = new Uint8Array(buf);

	for (i = 0; i < bytes.length; i++) {
		v8[i] = bytes[i];
	}

	return buf;
};

ArrayBuffer.combine = function (buf1, buf2) {
	var newBuf = new ArrayBuffer(buf1.byteLength + buf2.byteLength);

	var dst = new Int8Array(newBuf);
	var src;

	var i;

	src = new Int8Array(buf1);

	for (i = 0; i < buf1.byteLength; i++) {
		dst[i] = src[i];
	}

	src = new Int8Array(buf2);

	for (i = 0; i < buf2.byteLength; i++) {
		dst[i + buf1.byteLength] = src[i];
	}

	return dst;
};

})();
